import pytest

import tooltime.spec


examples = []
for equivalent_set in tooltime.spec.equivalent_sets['Timeperiod']:
    examples.extend(equivalent_set.items())


@pytest.mark.parametrize('example', examples)
def test_get_timeperiod_representation(example):
    actual_representation, value = example
    detected_representation = tooltime.detect_timeperiod_representation(value)
    assert actual_representation == detected_representation




import datetime
import typing

from .. import spec
from .. import timestamp_utils
from .. import timelength_utils
from . import timeperiod_identify


def create_timeperiod(
    start: spec.Timestamp = None,
    end: spec.Timestamp = None,
    length: spec.Timelength = None,
    to_representation: spec.TimeperiodRepresentation = None,
) -> spec.Timeperiod:
    """create Timeperiod

    ## Inputs
    - start: Timestamp
    - end: Timestamp
    - length: Timelength
    - to_reprsentation: str name of Timeperiod representation

    ## Returns
    - Timeperiod with specified representation
    """

    if to_representation is None:
        to_representation = 'TimeperiodMap'

    if to_representation == 'TimeperiodMap':
        return create_timeperiod_map(start=start, end=end, length=length)
    elif to_representation == 'TimeperiodPair':
        return create_timeperiod_pair(start=start, end=end, length=length)
    else:
        raise Exception(
            'unknown timeperiod representation: ' + str(to_representation)
        )


def create_timeperiod_pair(
    start: spec.Timestamp = None,
    end: spec.Timestamp = None,
    length: spec.Timelength = None,
) -> spec.TimeperiodPair:
    """create Timeperiod with representation TimeperiodPair

    ## Inputs
    - start: Timestamp
    - end: Timestamp
    - length: Timelength

    ## Returns
    - TimeperiodPair
    """
    start, end = compute_start_end(start=start, end=end, length=length)
    return (start, end)


def create_timeperiod_map(
    start: spec.Timestamp = None,
    end: spec.Timestamp = None,
    length: spec.Timelength = None,
) -> spec.TimeperiodMap:
    """create Timeperiod with representation TimeperiodMap

    ## Inputs
    - start: Timestamp
    - end: Timestamp
    - length: Timelength

    ## Returns
    - TimeperiodMap
    """
    start, end = compute_start_end(start=start, end=end, length=length)
    return {'start': start, 'end': end}


def compute_start_end(
    start: spec.Timestamp = None,
    end: spec.Timestamp = None,
    length: spec.Timelength = None,
) -> tuple[typing.Union[int, float], typing.Union[int, float]]:
    """return start and end given two of start, end, and length

    ## Inputs
    - start: Timestamp
    - end: Timestamp
    - length: Timelength

    ## Returns
    - TimeperiodPair
    """

    # convert inputs to seconds
    if start is not None:
        start = timestamp_utils.timestamp_to_seconds(start)
    if end is not None:
        end = timestamp_utils.timestamp_to_seconds(end)
    if length is None:
        length_seconds: typing.Union[int, float, None] = None
    elif isinstance(length, int):
        length_seconds = length
    elif isinstance(length, datetime.timedelta) or isinstance(length, str):
        length_seconds = timelength_utils.timelength_to_seconds_precise(length)

    # compute unknown bounds
    if end is not None and length_seconds is not None:
        start = end - length_seconds
    elif start is not None and length_seconds is not None:
        end = start + length_seconds
    elif start is not None and end is not None:
        pass
    else:
        raise Exception('must specify exactly 2 inputs')

    if type(start).__name__.startswith('int'):
        start = int(start)
    else:
        start = float(start)
    if type(end).__name__.startswith('int'):
        end = int(end)
    else:
        end = float(end)

    return (start, end)


def compute_timeperiod_start_end(
    timeperiod: spec.Timeperiod,
) -> tuple[typing.Union[int, float], typing.Union[int, float]]:
    """compute start and end seconds of a timeperiod"""

    if timeperiod_identify.is_timeperiod_pair(timeperiod):
        start, end = timeperiod
    elif timeperiod_identify.is_timeperiod_map(timeperiod):
        start = timeperiod['start']
        end = timeperiod['end']
    else:
        representation = timeperiod_identify.detect_timeperiod_representation(
            timeperiod
        )
        raise Exception(
            'unknown timeperiod representation: ' + str(representation)
        )

    start_numerical = timestamp_utils.timestamp_to_numerical(start)
    end_numerical = timestamp_utils.timestamp_to_numerical(end)

    return (start_numerical, end_numerical)




import pytest

import tooltime.spec


@pytest.mark.parametrize(
    'timelength_conversions', tooltime.spec.equivalent_sets['Timelength'],
)
def test_convert_timelengths(timelength_conversions):
    for from_representation, from_timelength in timelength_conversions.items():
        for to_representation, to_timelength in timelength_conversions.items():
            converted_timelength = tooltime.convert_timelength(
                timelength=from_timelength,
                from_representation=from_representation,
                to_representation=to_representation,
            )
            assert converted_timelength == to_timelength




import typing

from .. import spec
from .. import timestamp_utils
from . import timeperiod_convert


def print_timeperiod(
    timeperiod: spec.Timeperiod,
    timestamp_representation: spec.TimestampRepresentation = 'TimestampLabel',
    print_kwargs: dict[str, typing.Any] = None,
) -> None:
    """print Timeperiod

    ## Inputs
    - timeperiod: Timeperiod
    - timestamp_representation: str name of Timestamp representation to print
    - print_kwargs: kwargs passed to print()
    """

    start, end = timeperiod_convert.timeperiod_to_pair(timeperiod)
    start_label = timestamp_utils.convert_timestamp(
        start,
        to_representation=timestamp_representation,
    )
    end_label = timestamp_utils.convert_timestamp(
        end,
        to_representation=timestamp_representation,
    )

    if print_kwargs is None:
        print_kwargs = {}
    print('[' + str(start_label) + ', ' + str(end_label) + ']', **print_kwargs)




import pytest

import tooltime.spec


@pytest.mark.parametrize(
    'timefrequency_conversions',
    tooltime.spec.equivalent_sets['Timefrequency'],
)
def test_convert_timefrequencys(timefrequency_conversions):
    for (
        from_representation,
        from_timefrequency,
    ) in timefrequency_conversions.items():
        for (
            to_representation,
            to_timefrequency,
        ) in timefrequency_conversions.items():
            converted_timefrequency = tooltime.convert_timefrequency(
                timefrequency=from_timefrequency,
                from_representation=from_representation,
                to_representation=to_representation,
            )

            if to_representation != 'TimefrequencyFrequency':
                to_timefrequency = tooltime.timefrequency_to_frequency(
                    to_timefrequency
                )
                converted_timefrequency = tooltime.timefrequency_to_frequency(
                    converted_timefrequency
                )

            assert converted_timefrequency == to_timefrequency




from .. import timelength_utils
from . import timefrequency_identify


def convert_timefrequency(
    timefrequency,
    to_representation,
    from_representation=None,
):
    """convert Timefrequency to a new representation

    ## Inputs
    - timefrequency: Timefrequency
    - to_representation: str of Timefrequency representation of input timefrequency
    - from_representation: str of target Timefrequency representation

    ## Returns
    - Timefrequency in specified representation
    """

    # determine current representation
    if from_representation is None:
        from_representation = (
            timefrequency_identify.detect_timefrequency_representation(
                timefrequency
            )
        )

    # check whether conversion is required
    if to_representation == from_representation:
        return timefrequency

    # perform conversion
    if to_representation == 'TimefrequencyFrequency':
        return timefrequency_to_frequency(
            timefrequency, from_representation=from_representation
        )
    elif to_representation == 'TimefrequencyCountPer':
        return timefrequency_to_count_per(
            timefrequency, from_representation=from_representation
        )
    elif to_representation == 'TimefrequencyInterval':
        return timefrequency_to_interval(
            timefrequency, from_representation=from_representation
        )
    else:
        raise Exception(
            'unknown timeperiod representation: ' + str(to_representation)
        )


def timefrequency_to_frequency(timefrequency, from_representation=None):
    """convert Timefrequency to TimefrequencyFrequency representation

    ## Inputs
    - timefrequency: Timefrequency
    - from_representation: str representation name of input timefrequency

    ## Returns
    - TimefrequencyFrequency
    """

    if from_representation is None:
        from_representation = (
            timefrequency_identify.detect_timefrequency_representation(
                timefrequency
            )
        )

    if from_representation == 'TimefrequencyFrequency':
        return timefrequency
    elif from_representation == 'TimefrequencyCountPer':
        per_seconds = timelength_utils.timelength_to_seconds_precise(
            timefrequency['per']
        )
        return timefrequency['count'] / float(per_seconds)
    elif from_representation == 'TimefrequencyInterval':
        interval_seconds = timelength_utils.timelength_to_seconds_precise(
            timefrequency['interval']
        )
        return 1 / interval_seconds
    else:
        raise Exception(
            'unknown Timefrequency representation: ' + str(from_representation)
        )


def timefrequency_to_count_per(timefrequency, from_representation=None):
    """convert Timefrequency to TimefrequencyCountPer representation

    ## Inputs
    - timefrequency: Timefrequency
    - from_representation: str representation name of input timefrequency

    ## Returns
    - TimefrequencyCountPer
    """

    if from_representation is None:
        from_representation = (
            timefrequency_identify.detect_timefrequency_representation()
        )

    if from_representation == 'TimefrequencyFrequency':
        return {'count': timefrequency, 'per': '1s'}
    elif from_representation == 'TimefrequencyCountPer':
        return timefrequency
    elif from_representation == 'TimefrequencyInterval':
        return {'count': 1, 'per': timefrequency['interval']}
    else:
        raise Exception(
            'unknown Timefrequency representation: ' + str(from_representation)
        )


def timefrequency_to_interval(timefrequency, from_representation=None):
    """convert Timefrequency to TimefrequencyInterval representation

    ## Inputs
    - timefrequency: Timefrequency
    - from_representation: str representation name of input timefrequency

    ## Returns
    - TimefrequencyInterval
    """

    if from_representation is None:
        from_representation = (
            timefrequency_identify.detect_timefrequency_representation()
        )

    if from_representation == 'TimefrequencyFrequency':
        return {'interval': 1 / float(timefrequency)}
    elif from_representation == 'TimefrequencyCountPer':
        per_seconds = timelength_utils.timelength_to_seconds(
            timefrequency['per']
        )
        return {'interval': per_seconds / timefrequency['count']}
    elif from_representation == 'TimefrequencyInterval':
        return timefrequency
    else:
        raise Exception(
            'unknown Timefrequency representation: ' + str(from_representation)
        )




from .. import timelength_utils


def create_timefrequency(
    frequency=None, count=None, per=None, interval=None, to_representation=None,
):
    """create Timefrequency

    ## Inputs
    - frequency: TimeFrequency
    - count: int or float count
    - per: TimeLength
    - interval: TimeLength
    - to_representation: str name of Timefrequency representation

    ## Returns
    - Timeperiod with specified representation
    """

    if to_representation is None:
        to_representation = 'TimefrequencyFrequency'

    kwargs = dict(frequency=frequency, count=count, per=per, interval=interval)
    if to_representation == 'TimefrequencyFrequency':
        return create_timefrequency_frequency(**kwargs)
    elif to_representation == 'TimefrequencyCountPer':
        return create_timefrequency_count_per(**kwargs)
    elif to_representation == 'TimefrequencyInterval':
        return create_timefrequency_interval(**kwargs)
    else:
        raise Exception(
            'unknown timefrequency representation: ' + str(to_representation)
        )


def create_timefrequency_frequency(
    frequency=None, count=None, per=None, interval=None
):
    """create Timefrequency with representation TimefrequencyFrequency

    ## Inputs
    - frequency: TimeFrequency
    - count: int or float count
    - per: TimeLength
    - interval: TimeLength

    ## Returns
    - TimefrequencyFrequency
    """

    if frequency is not None:
        return frequency
    elif count is not None and per is not None:
        per_seconds = timelength_utils.timelength_to_seconds(per)
        return float(count) / float(per_seconds)
    elif interval is not None:
        interval_seconds = timelength_utils.timelength_to_seconds(interval)
        return 1 / float(interval_seconds)
    else:
        raise Exception('specify either frequency, interval, or count and per')


def create_timefrequency_count_per(
    frequency=None, count=None, per=None, interval=None,
):
    """create Timefrequency with representation TimefrequencyCountPer

    ## Inputs
    - frequency: TimeFrequency
    - count: int or float count
    - per: TimeLength
    - interval: TimeLength

    ## Returns
    - TimefrequencyCountPer
    """

    if frequency is not None:
        return {'count': None, 'per': None}
    if count is not None and per is not None:
        return {'count': count, 'per': per}
    elif interval is not None:
        return {'count': 1, 'per': interval}
    else:
        raise Exception('specify either frequency, interval, or count and per')


def create_timefrequency_interval(
    frequency=None, count=None, per=None, interval=None,
):
    """create Timefrequency with representation TimefrequencyInterval

    ## Inputs
    - frequency: TimeFrequency
    - count: int or float count
    - per: TimeLength
    - interval: TimeLength

    ## Returns
    - TimefrequencyInterval
    """

    if frequency is not None:
        return {'interval': 1 / float(frequency)}
    elif count is not None and per is not None:
        per_seconds = timelength_utils.timelength_to_seconds(per)
        return {'interval': per_seconds / float(count)}
    elif interval is not None:
        return {'interval': interval}
    else:
        raise Exception('specify either frequency, interval, or count and per')




import typing

from .. import spec
from .. import timefrequency_utils
from .. import timelength_utils
from . import timestamp_convert


def summarize_timestamps(
    timestamps: typing.Sequence[spec.Timestamp],
) -> spec.TimestampSummary:
    """create summary of timestamps

    ## Inputs
    - timestamps: iterable of Timestamp
    """

    timestamps_precise: typing.List[spec.TimestampSecondsPrecise] = []
    for timestamp in timestamps:
        precise = timestamp_convert.timestamp_to_seconds_precise(timestamp)
        timestamps_precise.append(precise)

    n_t = len(timestamps)
    summary: spec.TimestampSummary = {'n_t': n_t}

    if n_t == 1:
        summary['start'] = timestamps_precise[0]
        summary['end'] = timestamps_precise[0]

    elif n_t > 1:
        if timestamps_precise[-1] < timestamps_precise[0]:
            timestamps_precise = timestamps_precise[::-1]

        n_unique = len(set(timestamps_precise))
        resolution = timefrequency_utils.detect_resolution(timestamps_precise)
        start = timestamp_convert.timestamp_to_label(timestamps_precise[0])
        end = timestamp_convert.timestamp_to_label(timestamps_precise[-1])
        duration = timestamps_precise[-1] - timestamps_precise[0]
        duration_label = timelength_utils.timelength_seconds_to_clock_phrase(
            duration
        )
        n_large_outliers = resolution['outliers']['large'].shape[0]
        n_small_outliers = resolution['outliers']['small'].shape[0]
        n_outliers = n_large_outliers + n_small_outliers

        if resolution['median_dt'] > 0.0001:
            import numpy as np

            n_ideal_iter = np.arange(
                timestamps_precise[0],
                timestamps_precise[-1] + resolution['median_dt'],
                resolution['median_dt'],
            )
            n_ideal = list(n_ideal_iter)
            summary['n_missing'] = len(n_ideal) - len(timestamps_precise)
        else:
            summary['n_missing'] = None

        summary['n_unique'] = n_unique
        summary['resolution'] = resolution
        summary['start'] = start
        summary['end'] = end
        summary['duration'] = duration
        summary['duration_label'] = duration_label
        summary['n_large_outliers'] = n_large_outliers
        summary['n_small_outliers'] = n_small_outliers
        summary['n_outliers'] = n_outliers

    return summary


def print_timestamp_summary(
    *,
    timestamps: typing.List[spec.Timestamp] = None,
    summary: typing.Optional[spec.TimestampSummary] = None,
    indent: typing.Optional[str] = None,
    print_kwargs: dict[str, typing.Any] = None
):
    """print summary of timestamps

    - specify either timestamps or summary
    - timestamps should be ordered
        - todo: do not require ordering

    ## Inputs
    - timestamps: iterable of Timestamp
    - summary: dict summary created by summarize_timestamps()
    - indent: str indent of each line in summary
    - print_kwargs: kwargs passed to print()
    """

    # validate inputs
    if indent is None:
        indent = ''
    if print_kwargs is None:
        print_kwargs = {}

    # create summary if not provided
    if summary is None:
        if timestamps is None:
            raise Exception('must specify timestamps or summary')
        summary = summarize_timestamps(timestamps)

    n_t = summary['n_t']
    print(indent + 'n_t:', n_t, **print_kwargs)

    if n_t == 1:
        print(indent + 'timestamp:', summary['start'], **print_kwargs)

    elif n_t > 1:

        n_unique = summary['n_unique']
        resolution = summary['resolution']
        start = summary['start']
        end = summary['end']
        duration = summary['duration']
        duration_label = summary['duration_label']
        n_large_outliers = summary['n_large_outliers']
        n_small_outliers = summary['n_small_outliers']
        n_outliers = summary['n_outliers']
        n_missing = summary['n_missing']
        outlier_rtol = resolution['outliers']['outlier_rtol']
        mean_dt = timelength_utils.timelength_to_label(duration / (n_t - 1))

        print(indent + 'n_unique:', n_unique, **print_kwargs)
        print(indent + 'extent:')
        print(
            '    ' + indent + 'start:',
            start,
            '(' + ('%.14g' % summary['start']) + ')',
            **print_kwargs
        )
        print(
            '    ' + indent + 'end:  ',
            end,
            '(' + ('%.14g' % summary['end']) + ')',
            **print_kwargs
        )
        print(
            '    ' + indent + 'duration:',
            duration_label,
            '(' + ('%.14g' % duration) + ' s)',
            **print_kwargs
        )
        print(indent + 'resolution:', **print_kwargs)
        print(
            '    ' + indent + 'median_dt:', resolution['label'], **print_kwargs
        )
        print('    ' + indent + 'mean_dt:', mean_dt, **print_kwargs)
        print(
            '    ' + indent + 'missing timestamps:',
            n_missing,
            '(if median_dt maintained)',
            **print_kwargs
        )
        print('    ' + indent + 'outlier_dts:', n_outliers, **print_kwargs)
        print(
            '        ' + indent + 'small:',
            n_small_outliers,
            '       dt < median_dt / (1 + outlier_rtol)',
            **print_kwargs
        )
        print(
            '        ' + indent + 'large:',
            n_large_outliers,
            '       dt > median_dt * (1 + outlier_rtol)',
            **print_kwargs
        )
        print(
            '        ' + indent + '(outlier_rtol =',
            str(outlier_rtol) + ')',
            **print_kwargs
        )




import copy
import typing

from .. import spec


def get_base_units() -> dict[str, int]:
    """return mapping {label: seconds} for standard base units of time"""
    return {
        '1s': 1,
        '1m': 60,
        '1h': 60 * 60,
        '1d': 60 * 60 * 24,
        '1w': 60 * 60 * 24 * 7,
        '1M': 60 * 60 * 24 * 30,
        '1y': 60 * 60 * 24 * 365,
    }


def get_singular_unit_labels() -> dict[spec.SingularTimeUnit, str]:
    """return mapping {english_name: TimelengthLabel} for singular base units"""
    return {
        'second': '1s',
        'minute': '1m',
        'hour': '1h',
        'day': '1d',
        'week': '1w',
        'month': '1M',
        'year': '1y',
    }


def get_plural_unit_labels() -> dict[spec.PluralTimeUnit, str]:
    """return mapping {english_name: TimelengthLabel} for plural base units"""
    return {
        'seconds': '1s',
        'minutes': '1m',
        'hours': '1h',
        'days': '1d',
        'weeks': '1w',
        'months': '1M',
        'years': '1y',
    }


def get_unit_labels() -> dict[str, str]:
    """return mapping of singular and plural unit labels"""
    return dict(
        typing.cast(dict[str, str], get_singular_unit_labels()),
        **typing.cast(dict[str, str], get_plural_unit_labels())
    )


def unit_letters_to_names() -> dict[str, spec.SingularTimeUnit]:
    """return mapping {TimelengthLabel: english_name} for singular base units"""
    return {v[-1]: k for k, v in get_singular_unit_labels().items()}


def datetime_singular_unit_labels() -> dict[spec.DatetimeUnit, str]:
    """return mapping {english_name: TimelengthLabel} for singular base units"""
    return {
        'second': '1s',
        'minute': '1m',
        'hour': '1h',
        'day': '1d',
        'month': '1M',
        'year': '1y',
    }


def datetime_unit_letters_to_names() -> dict[str, spec.DatetimeUnit]:
    return {v[-1]: k for k, v in datetime_singular_unit_labels().items()}


def get_english_to_pandas_units() -> dict[str, str]:
    # https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html#timeseries-offset-aliases
    return {
        'year': 'YS',
        'years': 'YS',
        'month': 'MS',
        'months': 'MS',
        'day': 'D',
        'days': 'D',
        'hour': 'H',
        'hours': 'H',
        'minute': 'T',
        'minutes': 'T',
    }


def get_pandas_unit_to_english() -> dict[str, str]:
    return {v: k for k, v in get_english_to_pandas_units().items()}




import typing

from .. import exceptions
from .. import spec
from .. import timelength_utils


def detect_timefrequency_representation(
    timefrequency: typing.Any,
) -> spec.TimefrequencyRepresentation:
    """return str name of Timefrequency representation"""
    if is_timefrequency_frequency(timefrequency):
        return 'TimefrequencyFrequency'
    elif is_timefrequency_count_per(timefrequency):
        return 'TimefrequencyCountPer'
    elif is_timefrequency_interval(timefrequency):
        return 'TimefrequencyInterval'
    else:
        raise exceptions.RepresentationDetectionException(
            'could not detect Timefrequency representation: '
            + str(timefrequency)
        )


def is_timefrequency(timefrequency):
    """return bool of whether input is Timefrequency"""
    try:
        detect_timefrequency_representation(timefrequency)
        return True
    except exceptions.RepresentationDetectionException:
        return False


def is_timefrequency_frequency(timefrequency):
    """return bool of whether input is TimefrequencyFrequency"""
    return isinstance(timefrequency, (int, float))


def is_timefrequency_count_per(timefrequency):
    """return bool of whether input is TimefrequencyCountPer"""
    return (
        isinstance(timefrequency, dict)
        and len(timefrequency) == 2
        and 'count' in timefrequency
        and 'per' in timefrequency
        and isinstance(timefrequency['count'], (int, float))
        and timelength_utils.is_timelength(timefrequency['per'])
    )


def is_timefrequency_interval(timefrequency):
    """return bool of whether input is TimefrequencyInterval"""
    return (
        isinstance(timefrequency, dict)
        and len(timefrequency) == 1
        and 'interval' in timefrequency
        and timelength_utils.is_timelength(timefrequency['interval'])
    )




from .timeperiod_convert import *
from .timeperiod_crud import *
from .timeperiod_identify import *
from .timeperiod_introspect import *
from .timeperiod_overlap import *
from .timeperiod_standard import *




import datetime

from .. import spec


def floor_datetime(
    dt: datetime.datetime,
    precision: spec.DatetimeUnit,
) -> datetime.datetime:
    """take floor of datetime down to a given level of precision

    ## Inputs
    - dt: datetime object
    - precision: str name of precision unit to take floor to
    """

    if precision == 'year':
        remove: list[spec.DatetimeUnit] = [
            'month',
            'day',
            'hour',
            'minute',
            'second',
            'microsecond',
        ]
    elif precision == 'month':
        remove = ['day', 'hour', 'minute', 'second', 'microsecond']
    elif precision == 'day':
        remove = ['hour', 'minute', 'second', 'microsecond']
    elif precision == 'hour':
        remove = ['minute', 'second', 'microsecond']
    elif precision == 'minute':
        remove = ['second', 'microsecond']
    elif precision == 'second':
        remove = ['microsecond']
    elif precision == 'microsecond':
        return dt
    else:
        raise Exception('unknown precision: ' + str(precision))

    unit_values: dict[str, int] = {
        unit_name: get_unit_lowest_value(unit_name) for unit_name in remove
    }
    dt = dt.replace(tzinfo=dt.tzinfo, **unit_values)

    return dt


def get_unit_lowest_value(datetime_unit: spec.DatetimeUnit) -> int:
    """return lowest value for given unit"""

    lowest_values = {
        'month': 1,
        'day': 1,
        'hour': 0,
        'minute': 0,
        'second': 0,
        'microsecond': 0,
    }
    return lowest_values[datetime_unit]




import pytest

import tooltime.spec


@pytest.mark.parametrize(
    'timeperiod_conversions', tooltime.spec.equivalent_sets['Timeperiod'],
)
def test_convert_timeperiods(timeperiod_conversions):
    for from_representation, from_timeperiod in timeperiod_conversions.items():
        for to_representation, to_timeperiod in timeperiod_conversions.items():
            converted_timeperiod = tooltime.convert_timeperiod(
                timeperiod=from_timeperiod,
                from_representation=from_representation,
                to_representation=to_representation,
            )
            assert converted_timeperiod == to_timeperiod





from .. import spec
from . import timelength_convert


def create_timelength(
    seconds: spec.TimelengthSecondsRaw,
    to_representation: spec.TimelengthRepresentation = None,
) -> spec.Timelength:
    """create Timelength

    ## Inputs
    - seconds: int or float seconds
    - to_reprsentation: str name of Timelength representation

    ## Returns
    - Timelength with specified representation
    """

    if to_representation is None:
        to_representation = 'TimelengthSeconds'

    if to_representation == 'TimelengthSeconds':
        return create_timelength_seconds(seconds)
    elif to_representation == 'TimelengthSecondsPrecise':
        return create_timelength_seconds_precise(seconds)
    elif to_representation == 'TimelengthLabel':
        return create_timelength_label(seconds)
    elif to_representation == 'TimelengthPhrase':
        return create_timelength_phrase(seconds)
    elif to_representation == 'TimelengthClock':
        return create_timelength_clock(seconds)
    elif to_representation == 'TimelengthClockPhrase':
        return create_timelength_clock_phrase(seconds)
    elif to_representation == 'TimelengthTimedelta':
        return create_timelength_timedelta(seconds)
    else:
        raise Exception(
            'unknown timelength representation: ' + str(to_representation)
        )


def create_timelength_seconds(
    seconds: spec.TimelengthSecondsRaw,
) -> spec.TimelengthSeconds:
    """create Timelength with representation TimelengthSeconds"""
    return int(seconds)


def create_timelength_seconds_precise(
    seconds: spec.TimelengthSecondsRaw,
) -> spec.TimelengthSecondsPrecise:
    """create Timelength with representation TimelengthPrecise"""
    return float(seconds)


def create_timelength_label(
    seconds: spec.TimelengthSecondsRaw,
) -> spec.TimelengthLabel:
    """create Timelength with representation TimelengthLabel"""
    return timelength_convert.timelength_seconds_to_label(seconds)


def create_timelength_phrase(
    seconds: spec.TimelengthSecondsRaw,
) -> spec.TimelengthPhrase:
    """create Timelength with representation TimelengthPhrase"""
    return timelength_convert.timelength_seconds_to_phrase(seconds)


def create_timelength_clock(
    seconds: spec.TimelengthSecondsRaw,
) -> spec.TimelengthClock:
    """create Timelength with representation TimelengthClock"""
    return timelength_convert.timelength_seconds_to_clock(seconds)


def create_timelength_clock_phrase(
    seconds: spec.TimelengthSecondsRaw,
) -> spec.TimelengthClockPhrase:
    """create Timelength with representation TimelengthClockPhrase"""
    return timelength_convert.timelength_seconds_to_clock_phrase(seconds)


def create_timelength_timedelta(
    seconds: spec.TimelengthSecondsRaw,
) -> spec.TimelengthTimedelta:
    """create Timelength with representation TimelengthTimedelta"""
    return timelength_convert.timelength_seconds_to_timedelta(seconds)




import datetime

import pytest

import tooltime


# order: [block_unit, block_size, dt_value, block_start, block_end]
standard_timeperiod_test_tuples = [
    ['month', 4, 1, 1, 4],
    ['month', 4, 2, 1, 4],
    ['month', 4, 3, 1, 4],
    ['month', 4, 4, 1, 4],
    ['month', 4, 5, 5, 8],
    ['month', 4, 6, 5, 8],
    ['month', 4, 7, 5, 8],
    ['month', 4, 8, 5, 8],
    ['month', 4, 9, 9, 12],
    ['month', 4, 10, 9, 12],
    ['month', 4, 11, 9, 12],
    ['month', 4, 12, 9, 12],
    ['minute', 4, 0, 0, 3],
    ['minute', 4, 1, 0, 3],
    ['minute', 4, 2, 0, 3],
    ['minute', 4, 3, 0, 3],
    ['minute', 4, 4, 4, 7],
    ['minute', 4, 5, 4, 7],
    ['minute', 4, 6, 4, 7],
    ['minute', 4, 7, 4, 7],
    ['minute', 4, 8, 8, 11],
    ['minute', 4, 9, 8, 11],
    ['minute', 4, 10, 8, 11],
    ['minute', 4, 11, 8, 11],
    ['minute', 4, 12, 12, 15],
    ['minute', 4, 55, 52, 55],
    ['minute', 4, 56, 56, 59],
    ['minute', 4, 57, 56, 59],
    ['minute', 4, 58, 56, 59],
    ['minute', 4, 59, 56, 59],
    ['second', 4, 0, 0, 3],
    ['second', 4, 1, 0, 3],
    ['second', 4, 2, 0, 3],
    ['second', 4, 3, 0, 3],
    ['second', 4, 4, 4, 7],
]


@pytest.mark.parametrize('test_tuple', standard_timeperiod_test_tuples)
def test_get_standard_timeperiod(test_tuple):
    block_unit, block_size, dt_value, block_start, block_end = test_tuple

    dt = datetime.datetime.now(datetime.timezone.utc)
    dt = tooltime.floor_datetime(dt, block_unit)
    dt = dt.replace(**{block_unit: dt_value})
    timeperiod = tooltime.get_standard_timeperiod(
        timestamp=dt.timestamp(), block_unit=block_unit, block_size=block_size,
    )
    from_dt = tooltime.timestamp_to_datetime(timeperiod['start'])
    to_dt = tooltime.timestamp_to_datetime(timeperiod['end'])
    assert getattr(from_dt, block_unit) == block_start
    assert getattr(to_dt, block_unit) == block_end




import pytest

import tooltime.spec


@pytest.mark.parametrize(
    'timestamp_conversions', tooltime.spec.equivalent_sets['Timestamp'],
)
def test_convert_timestamps(timestamp_conversions):
    for from_representation, from_timestamp in timestamp_conversions.items():
        for to_representation, to_timestamp in timestamp_conversions.items():
            converted_timestamp = tooltime.convert_timestamp(
                timestamp=from_timestamp,
                from_representation=from_representation,
                to_representation=to_representation,
            )
            if to_representation == 'TimestampDatetime':
                assert (
                    converted_timestamp.timestamp() == to_timestamp.timestamp()
                )
            else:
                assert converted_timestamp == to_timestamp




import typing

from typing_extensions import TypeGuard

from .. import spec
from .. import exceptions
from .. import timestamp_utils


def detect_timeperiod_representation(
    timeperiod: spec.Timeperiod,
) -> spec.TimeperiodRepresentation:
    """return str name of Timeperiod representation"""
    if is_timeperiod_map(timeperiod):
        return 'TimeperiodMap'
    elif is_timeperiod_pair(timeperiod):
        return 'TimeperiodPair'
    else:
        raise exceptions.RepresentationDetectionException(
            'could not detect timeperiod representation: ' + str(timeperiod)
        )


def is_timeperiod(timeperiod: typing.Any) -> TypeGuard[spec.Timeperiod]:
    """return bool of whether input is Timeperiod"""
    try:
        detect_timeperiod_representation(timeperiod)
        return True
    except Exception:
        return False


def is_timeperiod_map(
    timeperiod: typing.Any,
) -> TypeGuard[spec.TimeperiodMap]:
    """return bool of whether input is TimeperiodMap"""
    return (
        isinstance(timeperiod, dict)
        and len(timeperiod) == 2
        and 'start' in timeperiod
        and 'end' in timeperiod
        and timestamp_utils.is_timestamp(timeperiod['start'])
        and timestamp_utils.is_timestamp(timeperiod['end'])
    )


def is_timeperiod_pair(
    timeperiod: typing.Any,
) -> TypeGuard[spec.TimeperiodPair]:
    """return bool of whether input is TimeperiodPair"""
    return (
        isinstance(timeperiod, tuple)
        and len(timeperiod) == 2
        and timestamp_utils.is_timestamp(timeperiod[0])
        and timestamp_utils.is_timestamp(timeperiod[1])
    )




import typing

from .. import spec
from . import timeperiod_identify


@typing.overload
def convert_timeperiod(
    timeperiod: spec.Timeperiod,
    to_representation: typing.Literal['TimeperiodPair'],
    from_representation: typing.Optional[spec.TimeperiodRepresentation],
) -> spec.TimeperiodPair:
    ...


@typing.overload
def convert_timeperiod(
    timeperiod: spec.Timeperiod,
    to_representation: typing.Literal['TimeperiodMap'],
    from_representation: typing.Optional[spec.TimeperiodRepresentation],
) -> spec.TimeperiodMap:
    ...


def convert_timeperiod(
    timeperiod: spec.Timeperiod,
    to_representation: spec.TimeperiodRepresentation,
    from_representation: typing.Optional[spec.TimeperiodRepresentation] = None,
) -> spec.Timeperiod:
    """convert Timeperiod to a new representation

    ## Inputs
    - timeperiod: Timeperiod
    - to_representation: str of Timeperiod representation of input timeperiod
    - from_representation: str of target Timeperiod representation

    ## Returns
    - Timeperiod in specified representation
    """

    # determine current representation
    if from_representation is None:
        from_representation = (
            timeperiod_identify.detect_timeperiod_representation(timeperiod)
        )

    # check whether conversion is required
    if to_representation == from_representation:
        return timeperiod

    # perform conversion
    if to_representation == 'TimeperiodPair':
        return timeperiod_to_pair(
            timeperiod, from_representation=from_representation
        )
    elif to_representation == 'TimeperiodMap':
        return timeperiod_to_map(
            timeperiod, from_representation=from_representation
        )
    else:
        raise Exception(
            'unknown timeperiod representation: ' + str(to_representation)
        )


def timeperiod_to_pair(
    timeperiod: spec.Timeperiod,
    from_representation: spec.TimeperiodRepresentation = None,
) -> spec.TimeperiodPair:
    """convert Timeperiod to TimeperiodPair

    ## Inputs
    - timeperiod: Timeperiod
    - from_representation: str representation name of input timeperiod

    ## Returns
    - TimeperiodPair
    """

    if timeperiod_identify.is_timeperiod_map(timeperiod):
        return (timeperiod['start'], timeperiod['end'])
    elif timeperiod_identify.is_timeperiod_pair(timeperiod):
        return timeperiod
    else:
        from_representation = (
            timeperiod_identify.detect_timeperiod_representation(timeperiod)
        )
        raise Exception(
            'unknown Timeperiod representation: ' + str(from_representation)
        )


def timeperiod_to_map(
    timeperiod: spec.Timeperiod,
    from_representation: spec.TimeperiodRepresentation = None,
) -> spec.TimeperiodMap:
    """convert Timeperiod to TimeperiodMap

    ## Inputs
    - timeperiod: Timeperiod
    - from_representation: str representation name of input timeperiod

    ## Returns
    - TimeperiodMap
    """

    if timeperiod_identify.is_timeperiod_pair(timeperiod):
        start, end = timeperiod
        return {'start': start, 'end': end}
    elif timeperiod_identify.is_timeperiod_map(timeperiod):
        return timeperiod
    else:
        from_representation = (
            timeperiod_identify.detect_timeperiod_representation(timeperiod)
        )
        raise Exception(
            'unknown Timeperiod representation: ' + str(from_representation)
        )




import datetime
import typing
from typing_extensions import TypeGuard

from .. import exceptions
from .. import spec
from . import timestamp_convert


def detect_timestamp_representation(
    timestamp: spec.Timestamp,
) -> spec.TimestampRepresentation:
    """return str name of Timestamp representation"""
    if is_timestamp_seconds(timestamp):
        return 'TimestampSeconds'
    elif is_timestamp_seconds_precise(timestamp):
        return 'TimestampSecondsPrecise'
    elif is_timestamp_label(timestamp):
        return 'TimestampLabel'
    elif is_timestamp_iso(timestamp):
        return 'TimestampISO'
    elif is_timestamp_datetime(timestamp):
        return 'TimestampDatetime'
    else:
        raise exceptions.RepresentationDetectionException(
            'could not detect Timestamp representation: ' + str(timestamp)
        )


def is_timestamp(timestamp: typing.Any) -> TypeGuard[spec.Timestamp]:
    """return bool of whether input is Timestamp"""
    try:
        detect_timestamp_representation(timestamp)
        return True
    except Exception:
        return False


def is_timestamp_seconds(
    timestamp: typing.Any,
) -> TypeGuard[spec.TimestampSeconds]:
    """return bool of whether input is TimestampSeconds"""
    return isinstance(timestamp, int) or type(timestamp).__name__ in [
        'int16',
        'int32',
        'int64',
    ]


def is_timestamp_seconds_precise(
    timestamp: typing.Any,
) -> TypeGuard[spec.TimestampSecondsPrecise]:
    """return bool of whether input is TimestampSecondsPrecise"""
    return isinstance(timestamp, float)


def is_timestamp_label(
    timestamp: typing.Any,
) -> TypeGuard[spec.TimestampLabel]:
    """return bool of whether input is TimestampLabel"""
    try:
        timestamp_convert.timestamp_label_to_seconds(
            typing.cast(spec.TimestampLabel, timestamp)
        )
        return True
    except Exception:
        return False


def is_timestamp_iso(
    timestamp: typing.Any,
) -> TypeGuard[spec.TimestampISO]:
    """return bool of whether input is TimestampISO"""
    try:
        timestamp_convert.timestamp_iso_to_seconds(
            typing.cast(spec.TimestampISO, timestamp)
        )
        return True
    except Exception:
        return False


def is_timestamp_datetime(
    timestamp: typing.Any,
) -> TypeGuard[spec.TimestampDatetime]:
    """return bool of whether input is TimestampDatetime"""
    return isinstance(timestamp, datetime.datetime)




import datetime
import math
import typing

from .. import spec
from . import timelength_units
from . import timelength_identify


#
# # general conversions
#


@typing.overload
def convert_timelength(
    timelength: spec.Timelength,
    to_representation: typing.Literal['TimelengthSeconds'],
    from_representation: typing.Optional[spec.TimelengthRepresentation],
) -> spec.TimelengthSeconds:
    ...


@typing.overload
def convert_timelength(
    timelength: spec.Timelength,
    to_representation: typing.Literal['TimelengthSecondsPrecise'],
    from_representation: typing.Optional[spec.TimelengthRepresentation],
) -> spec.TimelengthSecondsPrecise:
    ...


@typing.overload
def convert_timelength(
    timelength: spec.Timelength,
    to_representation: typing.Literal['TimelengthLabel'],
    from_representation: typing.Optional[spec.TimelengthRepresentation],
) -> spec.TimelengthLabel:
    ...


@typing.overload
def convert_timelength(
    timelength: spec.Timelength,
    to_representation: typing.Literal['TimelengthClock'],
    from_representation: typing.Optional[spec.TimelengthRepresentation],
) -> spec.TimelengthClock:
    ...


@typing.overload
def convert_timelength(
    timelength: spec.Timelength,
    to_representation: typing.Literal['TimelengthPhrase'],
    from_representation: typing.Optional[spec.TimelengthRepresentation],
) -> spec.TimelengthPhrase:
    ...


@typing.overload
def convert_timelength(
    timelength: spec.Timelength,
    to_representation: typing.Literal['TimelengthClockPhrase'],
    from_representation: typing.Optional[spec.TimelengthRepresentation],
) -> spec.TimelengthClockPhrase:
    ...


@typing.overload
def convert_timelength(
    timelength: spec.Timelength,
    to_representation: typing.Literal['TimelengthTimedelta'],
    from_representation: typing.Optional[spec.TimelengthRepresentation],
) -> spec.TimelengthTimedelta:
    ...


def convert_timelength(
    timelength: spec.Timelength,
    to_representation: spec.TimelengthRepresentation,
    from_representation: typing.Optional[spec.TimelengthRepresentation] = None,
) -> spec.Timelength:
    """convert Timelength to a new representation

    ## Inputs
    - timelength: Timelength
    - to_representation: str of Timelength representation of input timelength
    - from_representation: str of target Timelength representation

    ## Returns
    - Timelength in specified representation
    """

    # detect timelength type
    if from_representation is None:
        from_representation = (
            timelength_identify.detect_timelength_representation(timelength)
        )

    # return if already in target type
    if from_representation == to_representation:
        return timelength

    # convert to seconds
    if timelength_identify.is_timelength_seconds(timelength):
        timelength_seconds: spec.TimelengthSecondsRaw = timelength
    elif timelength_identify.is_timelength_seconds_precise(timelength):
        timelength_seconds = timelength
    elif timelength_identify.is_timelength_label(timelength):
        timelength_seconds = timelength_label_to_seconds(timelength)
    elif timelength_identify.is_timelength_clock(timelength):
        timelength_seconds = timelength_clock_to_seconds(timelength)
    elif timelength_identify.is_timelength_phrase(timelength):
        timelength_seconds = timelength_phrase_to_seconds(timelength)
    elif timelength_identify.is_timelength_clock_phrase(timelength):
        timelength_seconds = timelength_clock_phrase_to_seconds(timelength)
    elif timelength_identify.is_timelength_timedelta(timelength):
        timelength_seconds = timelength_timedelta_to_seconds(timelength)
    else:
        raise Exception(
            'unknown timelength_representation: ' + str(from_representation)
        )

    # convert to target type
    if to_representation == 'TimelengthSeconds':
        return int(timelength_seconds)
    elif to_representation == 'TimelengthSecondsPrecise':
        return float(timelength_seconds)
    elif to_representation == 'TimelengthLabel':
        return timelength_seconds_to_label(timelength_seconds)
    elif to_representation == 'TimelengthClock':
        return timelength_seconds_to_clock(timelength_seconds)
    elif to_representation == 'TimelengthPhrase':
        return timelength_seconds_to_phrase(timelength_seconds)
    elif to_representation == 'TimelengthClockPhrase':
        return timelength_seconds_to_clock_phrase(timelength_seconds)
    elif to_representation == 'TimelengthTimedelta':
        return timelength_seconds_to_timedelta(timelength_seconds)
    else:
        raise Exception(
            'unknown timelength_representation: ' + str(to_representation)
        )


#
# # functions with target representation specified
#


def timelength_to_seconds(
    timelength: spec.Timelength,
    from_representation: spec.TimelengthRepresentation = None,
) -> spec.TimelengthSeconds:
    """convert Timelength to TimelengthSeconds

    ## Inputs
    - timelength: Timelength
    - from_representation: str representation name of input timelength

    ## Returns
    - TimelengthSeconds timelength
    """
    return convert_timelength(
        timelength=timelength,
        to_representation='TimelengthSeconds',
        from_representation=from_representation,
    )


def timelength_to_seconds_precise(
    timelength: spec.Timelength,
    from_representation: spec.TimelengthRepresentation = None,
) -> spec.TimelengthSecondsPrecise:
    """convert Timelength to TimelengthSecondsPrecise

    ## Inputs
    - timelength: Timelength
    - from_representation: str representation name of input timelength

    ## Returns
    - TimelengthSecondsPrecise timelength
    """
    return convert_timelength(
        timelength=timelength,
        to_representation='TimelengthSecondsPrecise',
        from_representation=from_representation,
    )


def timelength_to_label(
    timelength: spec.Timelength,
    from_representation: spec.TimelengthRepresentation = None,
) -> spec.TimelengthLabel:
    """convert Timelength to TimelengthLabel

    ## Inputs
    - timelength: Timelength
    - from_representation: str representation name of input timelength

    ## Returns
    - TimelengthLabel timelength
    """
    return convert_timelength(
        timelength=timelength,
        to_representation='TimelengthLabel',
        from_representation=from_representation,
    )


def timelength_to_phrase(
    timelength: spec.Timelength,
    from_representation: spec.TimelengthRepresentation = None,
) -> spec.TimelengthPhrase:
    """convert Timelength to TimelengthPhrase

    ## Inputs
    - timelength: Timelength
    - from_representation: str representation name of input timelength

    ## Returns
    - TimelengthPhrase timelength
    """
    return convert_timelength(
        timelength=timelength,
        to_representation='TimelengthPhrase',
        from_representation=from_representation,
    )


def timelength_to_clock(
    timelength: spec.Timelength,
    from_representation: spec.TimelengthRepresentation = None,
) -> spec.TimelengthClock:
    """convert Timelength to TimelengthClock

    ## Inputs
    - timelength: Timelength
    - from_representation: str representation name of input timelength

    ## Returns
    - TimelengthClock timelength
    """
    return convert_timelength(
        timelength=timelength,
        to_representation='TimelengthClock',
        from_representation=from_representation,
    )


def timelength_to_clock_phrase(
    timelength: spec.Timelength,
    from_representation: spec.TimelengthRepresentation = None,
) -> spec.TimelengthClockPhrase:
    """convert Timelength to TimelengthClockPhrase

    ## Inputs
    - timelength: Timelength
    - from_representation: str representation name of input timelength

    ## Returns
    - TimelengthClockPhrase timelength
    """
    return convert_timelength(
        timelength=timelength,
        to_representation='TimelengthClockPhrase',
        from_representation=from_representation,
    )


#
# # specific conversion functions, from seconds
#


def timelength_seconds_to_label(
    timelength_seconds: spec.TimelengthSecondsRaw,
    base_only: bool = False,
    fuzzy_tolerance: typing.SupportsFloat = None,
    base_unit: str = None,
) -> spec.TimelengthLabel:
    """convert seconds to TimelengthLabel

    - matches integer multiples of base units
        - see get_base_units() for base units

    ## Inputs
    - seconds: int or float number of seconds
    - base_only: bool of whether to match only to base units
    - fuzzy_tolerance: float of how close seconds much be to a mathcing label
        - 0.1 -> '1m' would match any seconds in range [54, 66]
        - 0.1 -> '5m' would match any seconds in range [270, 330]
    - base_unit: str of base unit
    """

    seconds = timelength_seconds
    if type(seconds).__name__ in ['float', 'float64', 'float32', 'float16']:
        int_seconds = int(seconds)
        if math.isclose(int_seconds, seconds):
            seconds = int_seconds

        elif int_seconds > 5.0:
            seconds = int_seconds

    # collect possible base units
    base_units = timelength_units.get_base_units()
    if base_unit is not None:
        if base_unit not in base_units:
            raise Exception('invalid base unit: ' + str(base_unit))
        candidates = {base_unit: base_units[base_unit]}
    else:
        candidates = base_units

    # attempt matches to candidate base units in descending order
    descending = sorted(candidates.items(), key=lambda item: -item[-1])
    for base_label, base_seconds in descending:

        # attempt match to base unit
        if base_only:
            if math.isclose(seconds, base_seconds):
                return base_label
            else:
                continue

        # attempt exact match of base unit multiple
        quotient = seconds / base_seconds
        quotient_as_int = round(quotient)
        if math.isclose(quotient, quotient_as_int):
            unit_count = quotient_as_int
            unit_letter = base_label[-1]
            break

        # attempt fuzzy match of base unit multiple
        if fuzzy_tolerance is not None:
            round_quotient = round(quotient)
            factor = round_quotient * base_seconds
            upper_bound = (1 + spec.to_numeric(fuzzy_tolerance)) * factor
            lower_bound = (1 - spec.to_numeric(fuzzy_tolerance)) * factor
            if lower_bound <= seconds and seconds <= upper_bound:
                unit_count = round_quotient
                unit_letter = base_label[-1]
                break
    else:
        raise Exception('could not seconds convert to label')

    # create label
    label = str(unit_count) + unit_letter

    return label


def timelength_seconds_to_clock(
    timelength_seconds: spec.TimelengthSecondsRaw,
) -> spec.TimelengthClock:
    """convert seconds to TimelengthClock"""
    base_units = timelength_units.get_base_units()
    n_days = math.floor(timelength_seconds / base_units['1d'])
    n_seconds_remaining = timelength_seconds % base_units['1d']
    clock = str(datetime.timedelta(seconds=n_seconds_remaining))
    if n_days > 0:
        clock = str(n_days) + ':' + clock
    return clock


def timelength_seconds_to_phrase(
    timelength_seconds: spec.TimelengthSecondsRaw,
) -> spec.TimelengthPhrase:
    """convert seconds to TimelengthPhrase"""

    unit_names = [
        'years',
        'days',
        'hours',
        'minutes',
        'seconds',
    ]

    # compute count for each unit
    remaining = timelength_seconds
    base_units = timelength_units.get_base_units()
    unit_names_to_labels = timelength_units.get_unit_labels()
    unit_counts: list[typing.Union[int, float]] = []
    for unit_name in unit_names:
        unit_seconds = base_units[unit_names_to_labels[unit_name]]
        unit_count: typing.Union[int, float] = math.floor(
            remaining / unit_seconds
        )
        unit_counts.append(unit_count)
        remaining = remaining % unit_seconds
    if not math.isclose(remaining, 0):
        unit_counts[-1] += remaining

    # assemble pieces from unit counts
    pieces: list[str] = []
    for unit_count, unit_name in zip(unit_counts, unit_names):
        if unit_count > 0:
            piece = str(unit_count) + ' ' + unit_name
            pieces.append(piece)

    # assemble pieces into phrase
    if len(pieces) == 0:
        phrase = '0 ' + unit_names[-1]
    else:
        phrase = ', '.join(pieces)

    return phrase


def timelength_seconds_to_clock_phrase(
    timelength_seconds: spec.TimelengthSecondsRaw,
) -> spec.TimelengthClockPhrase:
    """convert seconds to TimelengthClockPhrase"""

    base_units = timelength_units.get_base_units()
    n_years = math.floor(timelength_seconds / base_units['1y'])
    n_years_remainder = timelength_seconds % base_units['1y']
    n_days = math.floor(n_years_remainder / base_units['1d'])
    n_days_remainder = n_years_remainder % base_units['1d']

    phrase_pieces = []
    if n_years > 0:
        phrase_pieces.append(str(n_years) + ' years')
    if n_days > 0:
        phrase_pieces.append(str(n_days) + ' days')
    if n_days_remainder > 0:
        if type(n_days_remainder).__name__ == 'int64':
            n_days_remainder = int(n_days_remainder)
        phrase_pieces.append(str(datetime.timedelta(seconds=n_days_remainder)))

    phrase = ', '.join(phrase_pieces)

    return phrase


def timelength_seconds_to_timedelta(
    timelength_seconds: spec.TimelengthSecondsRaw,
) -> spec.TimelengthTimedelta:
    """convert seconds to TimelengthTimedelta"""
    return datetime.timedelta(seconds=timelength_seconds)


#
# # to seconds
#


def timelength_label_to_seconds(
    timelength_label: spec.TimelengthLabel,
) -> spec.TimelengthSeconds:
    """convert TimelengthLabel to seconds"""
    number = int(timelength_label[:-1])
    letter = timelength_label[-1]
    base_units = timelength_units.get_base_units()
    base_seconds = base_units['1' + letter]
    seconds = number * base_seconds
    return seconds


def timelength_clock_to_seconds(
    timelength_clock: spec.TimelengthClock,
) -> spec.TimelengthSecondsRaw:
    """convert TimelengthClock to seconds"""

    pieces = timelength_clock.split(':')
    piece_numbers = [spec.str_to_numeric(piece) for piece in pieces]
    if len(pieces) == 3:
        days: typing.Union[int, float] = 0
        hours: typing.Union[int, float] = piece_numbers[0]
        minutes: typing.Union[int, float] = piece_numbers[1]
        seconds: typing.Union[int, float] = piece_numbers[2]
    elif len(pieces) == 4:
        days = piece_numbers[0]
        hours = piece_numbers[1]
        minutes = piece_numbers[2]
        seconds = piece_numbers[3]
    else:
        raise Exception('unknown timelength clock format')

    base_units = timelength_units.get_base_units()
    return (
        base_units['1d'] * days
        + base_units['1h'] * hours
        + base_units['1m'] * minutes
        + base_units['1s'] * seconds
    )


def timelength_phrase_to_seconds(
    timelength_phrase: spec.TimelengthPhrase,
) -> spec.TimelengthSecondsRaw:
    """convert TimelengthPhrase to seconds"""

    bases = timelength_units.get_base_units()
    unit_names_to_labels = timelength_units.get_unit_labels()
    phrase_pieces = timelength_phrase.split(', ')

    seconds: typing.Union[int, float] = 0
    for piece in phrase_pieces:
        unit_count, unit_name = piece.split(' ')
        unit_count_numeric = spec.str_to_numeric(unit_count)
        seconds += unit_count_numeric * bases[unit_names_to_labels[unit_name]]

    return seconds


def timelength_clock_phrase_to_seconds(
    timelength_clock_phrase: spec.TimelengthClockPhrase,
) -> spec.TimelengthSecondsRaw:
    """convert TimelengthClockPhrase to seconds"""

    pieces = timelength_clock_phrase.split(', ')
    phrase = ', '.join(pieces[:-1])
    if phrase != '':
        phrase_seconds = timelength_phrase_to_seconds(phrase)
    else:
        phrase_seconds = 0
    clock_seconds = timelength_clock_to_seconds(pieces[-1])
    return phrase_seconds + clock_seconds


def timelength_timedelta_to_seconds(
    timelength_timedelta: spec.TimelengthTimedelta,
) -> spec.TimelengthSecondsRaw:
    """convert TimelengthTimedelta to seconds"""
    return timelength_timedelta.total_seconds()


#
# # special conversions
#


def timelength_to_pandas_timelength(
    timelength: spec.Timelength,
) -> spec.TimelengthPandas:
    timelength_label = timelength_to_label(timelength)
    number = timelength_label[:-1]
    unit_name = timelength_units.unit_letters_to_names()[timelength_label[-1]]
    pandas_unit = timelength_units.get_english_to_pandas_units()[unit_name]
    return number + pandas_unit




import datetime
import typing


#
# # timestamp
#
TimestampSecondsRaw = typing.SupportsFloat
TimestampSeconds = int
TimestampSecondsPrecise = float
TimestampLabel = str
TimestampISO = str
TimestampDatetime = datetime.datetime
Timestamp = typing.Union[
    TimestampSeconds,
    TimestampSecondsPrecise,
    TimestampLabel,
    TimestampISO,
    TimestampDatetime,
]
TimestampRepresentation = typing.Literal[
    'TimestampSeconds',
    'TimestampSecondsPrecise',
    'TimestampLabel',
    'TimestampISO',
    'TimestampDatetime',
]
TimestampStrRepresentation = typing.Literal[
    'TimestampLabel',
    'TimestampISO',
]


TimestampSummary = dict[str, typing.Any]


#
# # timelength
#

TimelengthSecondsRaw = typing.Union[int, float]
TimelengthSeconds = int
TimelengthSecondsPrecise = float
TimelengthLabel = str
TimelengthClock = str
TimelengthPhrase = str
TimelengthClockPhrase = str
TimelengthTimedelta = datetime.timedelta
Timelength = typing.Union[
    TimelengthSeconds,
    TimelengthSecondsPrecise,
    TimelengthLabel,
    TimelengthClock,
    TimelengthPhrase,
    TimelengthClockPhrase,
    TimelengthTimedelta,
]
TimelengthRepresentation = typing.Literal[
    'TimelengthSeconds',
    'TimelengthSecondsPrecise',
    'TimelengthLabel',
    'TimelengthClock',
    'TimelengthPhrase',
    'TimelengthClockPhrase',
    'TimelengthTimedelta',
]

TimelengthPandas = str


#
# # timeperiod
#


class TimeperiodMap(typing.TypedDict):
    start: Timestamp
    end: Timestamp


class TimeperiodMapSeconds(typing.TypedDict):
    start: TimestampSeconds
    end: TimestampSeconds


TimeperiodPair = tuple[Timestamp, Timestamp]
Timeperiod = typing.Union[TimeperiodMap, TimeperiodPair]

TimeperiodRepresentation = typing.Literal[
    'TimeperiodPair',
    'TimeperiodMap',
]


#
# # timefrequency
#

TimefreqeuncyFrequency = typing.Union[int, float]


class TimefrequencyCountPer(typing.TypedDict):
    count: typing.Union[int, float]
    per: Timestamp


class TimefrequencyInterval(typing.TypedDict):
    interval: Timestamp


Timefrequency = typing.Union[
    TimefreqeuncyFrequency,
    TimefrequencyCountPer,
    TimefrequencyInterval,
]

TimefrequencyRepresentation = typing.Literal[
    'TimefrequencyFrequency',
    'TimefrequencyCountPer',
    'TimefrequencyInterval',
]


#
# # datetime
#

DatetimeUnit = typing.Literal[
    'year',
    'month',
    'day',
    'hour',
    'minute',
    'second',
    'microsecond',
]

SingularTimeUnit = typing.Literal[
    'year',
    'month',
    'week',
    'day',
    'hour',
    'minute',
    'second',
    'microsecond',
]

PluralTimeUnit = typing.Literal[
    'years',
    'months',
    'weeks',
    'days',
    'hours',
    'minutes',
    'seconds',
    'microseconds',
]


#
# # functions
#


def to_numeric(value: typing.SupportsFloat) -> typing.Union[int, float]:
    if isinstance(value, typing.SupportsInt) and type(
        value
    ).__name__.startswith('int'):
        return int(value)
    else:
        return float(value)


def str_to_numeric(value: str) -> typing.Union[int, float]:
    try:
        return int(value)
    except Exception:
        return float(value)


#
# # old
#

contenttypes = {
    # Timestamp
    'Timestamp': [
        'TimestampSeconds',
        'TimestampSecondsPrecise',
        'TimestampLabel',
        'TimestampISO',
        'TimestampDatetime',
    ],
    'TimestampSeconds': 'Integer',
    'TimestampSecondsPrecise': 'Float',
    'TimestampLabel': 'Text',
    'TimestampISO': 'Text',
    'TimestampDatetime': 'Datetime',
    # Timelength
    'Timelength': [
        'TimelengthSeconds',
        'TimelengthSecondsPrecise',
        'TimelengthLabel',
        'TimelengthClock',
        'TimelengthPhrase',
        'TimelengthClockPhrase',
        'TimelengthTimedelta',
    ],
    'TimelengthSeconds': 'Integer',
    'TimelengthSecondsPrecise': 'Float',
    'TimelengthLabel': 'Text',
    'TimelengthClock': 'Text',
    'TimelengthPhrase': 'Text',
    'TimelengthClockPhrase': 'Text',
    'TimelengthTimedelta': 'Timedelta',
    # Timeperiod
    'Timeperiod': ['TimeperiodMap', 'TimeperiodPair'],
    'TimeperiodMap': {'start': 'Timestamp', 'end': 'Timestamp'},
    'TimeperiodPair': ['Timestamp', 'Timestamp'],
    # Timefrequency
    'Timefrequency': [
        'TimefrequencyFrequency',
        'TimefrequencyCountPer',
        'TimefrequencyInterval',
    ],
    'TimeFrequencyFrequency': 'Number',
    'TimeFrequencyCount': {'count': 'Number', 'per': 'Timelength'},
    'TimeFrequencyInterval': {'interval': 'Timelength'},
}


equivalent_sets = {
    'Timestamp': [
        {
            'TimestampSeconds': 1600000000,
            'TimestampSecondsPrecise': 1600000000.0,
            'TimestampLabel': '20200913_122640Z',
            'TimestampISO': '2020-09-13T12:26:40Z',
            'TimestampDatetime': datetime.datetime(
                second=40,
                minute=26,
                hour=12,
                day=13,
                month=9,
                year=2020,
                tzinfo=datetime.timezone.utc,
            ),
        },
    ],
    'Timelength': [
        {
            'TimelengthSeconds': 31,
            'TimelengthSecondsPrecise': 31.0,
            'TimelengthLabel': '31s',
            'TimelengthClock': '0:00:31',
            'TimelengthPhrase': '31 seconds',
            'TimelengthClockPhrase': '0:00:31',
            'TimelengthTimedelta': datetime.timedelta(seconds=31),
        },
        {
            'TimelengthSeconds': 86431,
            'TimelengthSecondsPrecise': 86431.0,
            'TimelengthLabel': '86431s',
            'TimelengthClock': '1:0:00:31',
            'TimelengthPhrase': '1 days, 31 seconds',
            'TimelengthClockPhrase': '1 days, 0:00:31',
            'TimelengthTimedelta': datetime.timedelta(days=1, seconds=31),
        },
    ],
    'Timeperiod': [
        {
            'TimeperiodMap': {'start': 1600000000, 'end': 1600000001},
            'TimeperiodPair': (1600000000, 1600000001),
        },
    ],
    'Timefrequency': [
        {
            'TimefrequencyFrequency': 5.0,
            'TimefrequencyCountPer': {'count': 5, 'per': '1s'},
            'TimefrequencyInterval': {'interval': 0.2},
        },
    ],
}




import datetime
import typing

from typing_extensions import TypeGuard

from .. import spec
from .. import exceptions
from . import timelength_units


def detect_timelength_representation(
    timelength: spec.Timelength,
) -> spec.TimelengthRepresentation:
    """return str name of Timelength representation"""
    if is_timelength_seconds(timelength):
        return 'TimelengthSeconds'
    elif is_timelength_seconds_precise(timelength):
        return 'TimelengthSecondsPrecise'
    elif is_timelength_label(timelength):
        return 'TimelengthLabel'
    elif is_timelength_clock(timelength):
        return 'TimelengthClock'
    elif is_timelength_phrase(timelength):
        return 'TimelengthPhrase'
    elif is_timelength_clock_phrase(timelength):
        return 'TimelengthClockPhrase'
    elif is_timelength_timedelta(timelength):
        return 'TimelengthTimedelta'
    else:
        raise exceptions.RepresentationDetectionException(
            'could not determine Timelength representation: ' + str(timelength)
        )


def is_timelength(timelength: typing.Any) -> TypeGuard[spec.Timelength]:
    """return bool of whether input is Timelength"""
    try:
        detect_timelength_representation(timelength)
        return True
    except exceptions.RepresentationDetectionException:
        return False


def is_timelength_seconds(
    timelength: typing.Any,
) -> TypeGuard[spec.TimelengthSeconds]:
    """return bool of whether input is TimelengthSeconds"""
    return isinstance(timelength, int)


def is_timelength_seconds_precise(
    timelength: typing.Any,
) -> TypeGuard[spec.TimelengthSecondsPrecise]:
    """return bool of whether input is TimelengthSecondsPrecise"""
    return isinstance(timelength, float)


def is_timelength_label(
    timelength: typing.Any,
) -> TypeGuard[spec.TimelengthLabel]:
    """return bool of whether input is TimelengthLabel"""

    if not isinstance(timelength, str) or len(timelength) < 2:
        return False

    try:
        int(timelength[:-1])
        letter = timelength[-1]
        return letter.isalnum()

    except ValueError:
        return False


def is_timelength_clock(
    timelength: typing.Any,
) -> TypeGuard[spec.TimelengthClock]:
    """return bool of whether input is TimelengthClock"""

    if not isinstance(timelength, str):
        return False
    numbers = timelength.split(':')
    try:
        for number in numbers[:-1]:
            int(number)
        float(numbers[-1])
        return True
    except ValueError:
        return False


def is_timelength_phrase(
    timelength: typing.Any,
) -> TypeGuard[spec.TimelengthPhrase]:
    """return bool of whether input is TimelengthPhrase"""

    if not isinstance(timelength, str):
        return False
    unit_names_to_labels = timelength_units.get_unit_labels()
    pieces = timelength.split(', ')
    try:
        for piece in pieces:
            amount, unit_name = piece.split(' ')
            float(amount)
            if unit_name not in unit_names_to_labels:
                return False
        return True
    except Exception:
        return False


def is_timelength_clock_phrase(
    timelength: typing.Any,
) -> TypeGuard[spec.TimelengthClockPhrase]:
    """return bool of whether input is TimelengthClockPhrase"""

    if not isinstance(timelength, str):
        return False
    pieces = timelength.split(', ')

    if ':' in pieces[-1]:
        clock = pieces[-1]
        if not is_timelength_clock(clock):
            return False
        phrase = ', '.join(pieces[:-1])
    else:
        phrase = ', '.join(pieces)

    return is_timelength_phrase(phrase)


def is_timelength_timedelta(
    timelength: typing.Any,
) -> TypeGuard[spec.TimelengthTimedelta]:
    """return bool of whether input is TimelengthTimedelta"""
    return isinstance(timelength, datetime.timedelta)





import setuptools


with open("README.md", "r", encoding="utf-8") as f:
    long_description = f.read()


setuptools.setup(
    name='tooltime',
    version='0.0.2',
    description='tooltime makes it easier to perform operations on time data',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/sslivkoff/tooltime',
    packages=setuptools.find_packages(),
    install_requires=[
        'numpy',
        'pandas',
    ],
    python_requires='>=3.6',
    package_data={
        'tooltime': ['py.typed'],
    }
)




import typing

from .. import spec
from .. import timelength_utils
from . import timeperiod_crud


def timeperiods_overlap(
    timeperiod_lhs: spec.Timeperiod, timeperiod_rhs: spec.Timeperiod
) -> bool:
    """return bool of whether timeperiods have any overlap"""
    start_lhs, end_lhs = timeperiod_crud.compute_timeperiod_start_end(
        timeperiod_lhs
    )
    start_rhs, end_rhs = timeperiod_crud.compute_timeperiod_start_end(
        timeperiod_rhs
    )
    return (start_lhs <= start_rhs <= end_lhs) or (
        start_lhs <= end_rhs <= end_lhs
    )


def timeperiod_contains(
    timeperiod: spec.Timeperiod,
    other_timeperiod: spec.Timeperiod,
) -> bool:
    """return bool of whether timeperiod contains other timeperiod"""
    start, end = timeperiod_crud.compute_timeperiod_start_end(timeperiod)
    other_start, other_end = timeperiod_crud.compute_timeperiod_start_end(
        other_timeperiod
    )
    return (start <= other_start) and (end >= other_end)


def create_superset_timeperiod(
    *timeperiods: spec.Timeperiod,
) -> spec.Timeperiod:
    """create Timeperiod that contains all input Timeperiods"""
    min_start = float('inf')
    max_end = float('-inf')
    for timeperiod in timeperiods:
        start, end = timeperiod_crud.compute_timeperiod_start_end(timeperiod)
        if start < min_start:
            min_start = start
        if end > max_end:
            max_end = end
    return (min_start, max_end)


def create_overlapping_timeperiod(
    timeperiod: spec.Timeperiod,
    trim_start_relative: typing.SupportsFloat = None,
    trim_end_relative: typing.SupportsFloat = None,
    trim_start_absolute: spec.Timelength = None,
    trim_end_absolute: spec.Timelength = None,
    extend_start_relative: typing.SupportsFloat = None,
    extend_end_relative: typing.SupportsFloat = None,
    extend_start_absolute: spec.Timelength = None,
    extend_end_absolute: spec.Timelength = None,
) -> spec.Timeperiod:
    """create copy of timeperiod with start or end trimmed or extended

    - can trim or end by relative or absolute amount

    ## Inputs
    - timeperiod: Timeperiod
    - trim_start_relative: float relative amount of length to trim at start
    - trim_end_relative: float relative amount of length to trim at end
    - trim_start_absolute: Timelength amount of time to trim at start
    - trim_end_absolute: Timelength amount of time to trim at end
    - extend_start_relative: float relative amount of length to extend at start
    - extend_end_relative: float relative amount of length to extend at end
    - extend_start_absolute: Timelength amount of time to extend at start
    - extend_end_absolute: Timelength amount of time to extend at end
    """

    start, end = timeperiod_crud.compute_timeperiod_start_end(timeperiod)
    length = end - start
    new_start = start
    new_end = end

    # trim boundaries
    if trim_start_relative is not None:
        new_start = new_start + spec.to_numeric(trim_start_relative) * length
    if trim_end_relative is not None:
        new_end = new_end - spec.to_numeric(trim_end_relative) * length
    if trim_start_absolute is not None:
        trim_start_seconds = timelength_utils.timelength_to_seconds(
            trim_start_absolute
        )
        new_start = new_start + trim_start_seconds
    if trim_end_absolute is not None:
        trim_end_seconds = timelength_utils.timelength_to_seconds(
            trim_end_absolute
        )
        new_end = new_end - trim_end_seconds

    # extend boundaries
    if extend_start_relative is not None:
        new_start = new_start - spec.to_numeric(extend_start_relative) * length
    if extend_end_relative is not None:
        new_end = new_end + spec.to_numeric(extend_end_relative) * length
    if extend_start_absolute is not None:
        extend_start_seconds = timelength_utils.timelength_to_seconds(
            extend_start_absolute
        )
        new_start = new_start - extend_start_seconds
    if extend_end_absolute is not None:
        extend_end_seconds = timelength_utils.timelength_to_seconds(
            extend_end_absolute
        )
        new_end = new_end + extend_end_seconds

    return (new_start, new_end)




import pytest

import tooltime.spec


examples = []
for equivalent_set in tooltime.spec.equivalent_sets['Timelength']:
    examples.extend(equivalent_set.items())


@pytest.mark.parametrize('example', examples)
def test_get_timelength_representation(example):
    actual_representation, value = example
    detected_representation = tooltime.detect_timelength_representation(value)

    if actual_representation == 'TimelengthClockPhrase':
        assert detected_representation in [
            'TimelengthClock',
            'TimelengthClockPhrase',
        ]
    else:
        assert actual_representation == detected_representation




import time
import typing

from .. import spec
from . import timestamp_convert


def create_timestamp(
    seconds: typing.SupportsFloat = None,
    representation: spec.TimestampRepresentation = 'TimestampSeconds',
) -> spec.Timestamp:
    """create Timestamp

    ## Inputs
    - seconds: int or float utc seconds
    - representation: str name of Timestamp representation

    ## Returns
    - Timestamp with specified representation
    """
    if representation == 'TimestampSeconds':
        return create_timestamp_seconds(seconds=seconds)
    elif representation == 'TimestampSecondsPrecise':
        return create_timestamp_seconds_precise(seconds=seconds)
    elif representation == 'TimestampLabel':
        return create_timestamp_label(seconds=seconds)
    elif representation == 'TimestampISO':
        return create_timestamp_iso(seconds=seconds)
    elif representation == 'TimestampDatetime':
        return create_timestamp_datetime(seconds=seconds)
    else:
        raise Exception('unknown representation: ' + str(representation))


def create_timestamp_seconds(
    seconds: typing.SupportsFloat = None,
) -> spec.TimestampSeconds:
    """create Timestamp with representation TimestampSeconds"""
    if seconds is None:
        seconds = time.time()
    return int(float(seconds))


def create_timestamp_seconds_precise(
    seconds: typing.SupportsFloat = None,
) -> spec.TimestampSecondsPrecise:
    """create Timestamp with representation TimestampSecondsPrecise"""
    if seconds is None:
        seconds = time.time()
    return float(seconds)


def create_timestamp_label(
    seconds: typing.SupportsFloat = None,
) -> spec.TimestampLabel:
    """create Timestamp with representation TimestampLabel"""
    if seconds is None:
        seconds = time.time()
    return timestamp_convert.timestamp_seconds_to_label(seconds)


def create_timestamp_iso(
    seconds: typing.SupportsFloat = None,
) -> spec.TimestampISO:
    """create Timestamp with representation TimestampISO"""
    if seconds is None:
        seconds = time.time()
    return timestamp_convert.timestamp_seconds_to_iso(seconds)


def create_timestamp_datetime(
    seconds: typing.SupportsFloat = None,
) -> spec.TimestampDatetime:
    """create Timestamp with representation TimestampDatetime"""
    if seconds is None:
        seconds = time.time()
    return timestamp_convert.timestamp_seconds_to_datetime(seconds)




import datetime
import math
import typing

from .. import spec
from .. import timelength_utils
from .. import timestamp_utils


def get_standard_timeperiod(
    timelength_label: spec.TimelengthLabel = None,
    *,
    block_unit: typing.Optional[spec.DatetimeUnit] = None,
    block_size: typing.Optional[int] = None,
    timestamp: typing.Optional[spec.Timestamp] = None,
    include_start: bool = True,
    include_end: bool = False,
    boundary_unit: spec.DatetimeUnit = 'second',
) -> spec.TimeperiodMapSeconds:
    """get standardized Timeperiod that contains a specific Timestamp

    ## Standardized Timeperiods
    - standardized boundaries are integer multiples of some block_unit
    - should specify either timelength_label or block_unit to define block unit
    - should specify block_size to define integer multiple, or use default of 1

    ## Example Usage
    timeperiod = tooltime.get_standard_timeperiod(
        timestamp=1600000000,
        block_size=5,
        block_unit='minutes',
    )
    tooltime.print_timeperiod(timeperiod)
    > [20200913_122500Z, 20200913_122959Z]

    ## Inputs
    - timelength_label: TimelengthLabel
    - block_unit: int size of block unit
    - block_size: str name of time unit
    - timestamp: Timestamp contained in timeperiod, default is now
    - include_start: bool of whether to include start boundary of timeperiod
    - include_end: bool of whether to include end boundary of timeperiod
    - boundary_unit: str name of boundary unit to be shaved off open intervals
    """

    if (timelength_label is not None) and (block_unit is not None):
        raise Exception('must specify either timelength_label or block_unit')
    elif (timelength_label is None) and (block_unit is not None):
        if block_size is None:
            block_size = 1
    elif (timelength_label is not None) and (block_unit is None):
        block_size = int(timelength_label[:-1])
        unit_letters_to_names = (
            timelength_utils.datetime_unit_letters_to_names()
        )
        block_unit = unit_letters_to_names[timelength_label[-1]]
    elif (timelength_label is None) and (block_unit is None):
        raise Exception()
    else:
        raise Exception()

    if block_unit is None:
        raise Exception('block_unit must be set')

    # create timestamp datetime
    if timestamp is not None:
        dt = timestamp_utils.timestamp_to_datetime(timestamp)
    else:
        dt = datetime.datetime.now(datetime.timezone.utc)

    # compute from_datetime
    current_unit = getattr(dt, block_unit)
    lowest_unit_value = timestamp_utils.get_unit_lowest_value(block_unit)
    current_block = math.floor((current_unit - lowest_unit_value) / block_size)
    from_unit = current_block * block_size + lowest_unit_value
    from_datetime = timestamp_utils.floor_datetime(dt, block_unit)
    kwargs: dict[str, int] = {str(block_unit): from_unit}
    from_datetime = from_datetime.replace(tzinfo=from_datetime.tzinfo, **kwargs)

    # compute to_datetime
    if block_unit == 'month':
        # datetime.timedelta does not support months
        to_month_raw = from_datetime.month + block_size
        to_month = to_month_raw % 12
        to_year = from_datetime.year + math.floor(to_month_raw / 12)
        to_datetime = from_datetime.replace(month=to_month, year=to_year)
    else:
        block_timedelta = datetime.timedelta(**{(block_unit + 's'): block_size})
        to_datetime = from_datetime + block_timedelta

    # trim boundaries
    if not include_start:
        timedelta = datetime.timedelta(**{(boundary_unit + 's'): 1})
        from_datetime = from_datetime + timedelta
    if not include_end:
        timedelta = datetime.timedelta(**{(boundary_unit + 's'): 1})
        to_datetime = to_datetime - timedelta

    # get timestamps
    start = int(from_datetime.timestamp())
    end = int(to_datetime.timestamp())

    return {'start': start, 'end': end}


def get_standard_intervals(
    interval_size: spec.Timelength,
    start_time: typing.Optional[spec.Timestamp] = None,
    end_time: typing.Optional[spec.Timestamp] = None,
    n_intervals: typing.Optional[int] = None,
    window_size: typing.Optional[spec.Timelength] = None,
) -> list[spec.TimestampSeconds]:
    """
    ## Valid Inputs
    - {start_time, end_time, interval_size}
    - {start_time, interval_size, {n_intervals or window_size}}
    - {end_time, interval_size, {n_intervals or window_size}}
    - {interval_size, {n_intervals or window_size}}
    - note that interval_size is always necessary
    - cannot specify {start_time, end_time, {n_intervals or window_size}} because that won't be standardized
    """

    # validate inputs
    if (
        start_time is not None
        and end_time is not None
        and n_intervals is not None
    ):
        raise Exception(
            'cannot specify {start_time, end_time, n_intervals} because that won\'t be standardized'
        )

    # use current time as default
    if start_time is None and end_time is None:
        end_time = timestamp_utils.create_timestamp()

    date_range_kwargs: dict[str, typing.Any] = {}

    # parse interval_size
    if interval_size is not None:
        date_range_kwargs[
            'freq'
        ] = timelength_utils.timelength_to_pandas_timelength(interval_size)

    # parse n_intervals
    if window_size is not None:
        window_length = timelength_utils.timelength_to_seconds(window_size)
        interval_length = timelength_utils.timelength_to_seconds(interval_size)
        n_intervals = int(window_length / interval_length)
    if n_intervals is not None:
        date_range_kwargs['periods'] = n_intervals

    # parse start time
    if start_time is not None:
        timeperiod = get_standard_timeperiod(
            timestamp=start_time,
            timelength_label=timelength_utils.timelength_to_label(
                interval_size
            ),
            include_end=True,
        )
        date_range_kwargs['start'] = timeperiod['start'] * 1000000000

    # parse end time
    if end_time is not None:
        timeperiod = get_standard_timeperiod(
            timestamp=end_time,
            timelength_label=timelength_utils.timelength_to_label(
                interval_size
            ),
            include_end=True,
        )
        date_range_kwargs['end'] = timeperiod['end'] * 1000000000

    # create intervals
    import pandas as pd

    intervals = pd.date_range(**date_range_kwargs)

    # extract timestamps
    timestamps = [int(interval.timestamp()) for interval in intervals]

    return timestamps




import datetime
import typing

from .. import spec
from . import timestamp_identify


time_format = '%Y%m%d_%H%M%SZ'
precise_time_format = time_format[:-1] + '%f' + time_format[-1]


#
# # general conversion
#


@typing.overload
def convert_timestamp(
    timestamp: spec.Timestamp,
    to_representation: typing.Literal['TimestampSeconds'],
    from_representation: typing.Optional[spec.TimestampRepresentation] = None,
) -> spec.TimestampSeconds:
    ...


@typing.overload
def convert_timestamp(
    timestamp: spec.Timestamp,
    to_representation: typing.Literal['TimestampSecondsPrecise'],
    from_representation: typing.Optional[spec.TimestampRepresentation] = None,
) -> spec.TimestampSecondsPrecise:
    ...


@typing.overload
def convert_timestamp(
    timestamp: spec.Timestamp,
    to_representation: typing.Literal['TimestampLabel'],
    from_representation: typing.Optional[spec.TimestampRepresentation] = None,
) -> spec.TimestampLabel:
    ...


@typing.overload
def convert_timestamp(
    timestamp: spec.Timestamp,
    to_representation: typing.Literal['TimestampISO'],
    from_representation: typing.Optional[spec.TimestampRepresentation] = None,
) -> spec.TimestampISO:
    ...


@typing.overload
def convert_timestamp(
    timestamp: spec.Timestamp,
    to_representation: typing.Literal['TimestampDatetime'],
    from_representation: typing.Optional[spec.TimestampRepresentation] = None,
) -> spec.TimestampDatetime:
    ...


def convert_timestamp(
    timestamp: spec.Timestamp,
    to_representation: spec.TimestampRepresentation,
    from_representation: typing.Optional[spec.TimestampRepresentation] = None,
) -> spec.Timestamp:
    """convert timestamp to a new representation

    ## Inputs
    - timestamp: Timestamp
    - to_representation: str of Timestamp representation of input timestamp
    - from_representation: str of target Timestamp representation

    ## Returns
    - Timestamp in specified representation
    """

    # determine current representation
    if from_representation is None:
        from_representation = (
            timestamp_identify.detect_timestamp_representation(timestamp)
        )

    # check if representation is required
    if from_representation == to_representation:
        return timestamp

    # convert to seconds
    if timestamp_identify.is_timestamp_seconds(timestamp):
        timestamp_seconds: spec.TimestampSecondsRaw = timestamp
    elif timestamp_identify.is_timestamp_seconds_precise(timestamp):
        timestamp_seconds = timestamp
    elif timestamp_identify.is_timestamp_label(timestamp):
        timestamp_seconds = timestamp_label_to_seconds(timestamp)
    elif timestamp_identify.is_timestamp_iso(timestamp):
        timestamp_seconds = timestamp_iso_to_seconds(timestamp)
    elif timestamp_identify.is_timestamp_datetime(timestamp):
        timestamp_seconds = timestamp_datetime_to_seconds(timestamp)
    else:
        raise Exception(
            'unknown timestamp representation: ' + str(from_representation)
        )

    # convert to target representation
    if to_representation == 'TimestampSeconds':
        return int(float(timestamp_seconds))
    elif to_representation == 'TimestampSecondsPrecise':
        return float(timestamp_seconds)
    elif to_representation == 'TimestampLabel':
        return timestamp_seconds_to_label(timestamp_seconds)
    elif to_representation == 'TimestampISO':
        return timestamp_seconds_to_iso(timestamp_seconds)
    elif to_representation == 'TimestampDatetime':
        return timestamp_seconds_to_datetime(timestamp_seconds)
    else:
        raise Exception(
            'unknown timestamp representation: ' + str(to_representation)
        )


#
# # functions with target representation specified
#


def timestamp_to_seconds(
    timestamp: spec.Timestamp,
    from_representation: spec.TimestampRepresentation = None,
) -> spec.TimestampSeconds:
    """convert timestamp to TimestampSeconds

    ## Inputs
    - timestamp: Timestamp
    - from_representation: str representation name of input timestamp

    ## Returns
    - TimestampSeconds timestamp
    """
    return convert_timestamp(
        timestamp,
        to_representation='TimestampSeconds',
        from_representation=from_representation,
    )


def timestamp_to_seconds_precise(
    timestamp: spec.Timestamp,
    from_representation: spec.TimestampRepresentation = None,
) -> spec.TimestampSecondsPrecise:
    """convert timestamp to TimestampSecondsPrecise

    ## Inputs
    - timestamp: Timestamp
    - from_representation: str representation name of input timestamp

    ## Returns
    - TimestampSecondsPrecise timestamp
    """
    return convert_timestamp(
        timestamp,
        to_representation='TimestampSecondsPrecise',
        from_representation=from_representation,
    )


def timestamp_to_label(
    timestamp: spec.Timestamp,
    from_representation: spec.TimestampRepresentation = None,
) -> spec.TimestampLabel:
    """convert timestamp to TimestampLabel

    ## Inputs
    - timestamp: Timestamp
    - from_representation: str representation name of input timestamp

    ## Returns
    - TimestampLabel timestamp
    """
    return convert_timestamp(
        timestamp,
        to_representation='TimestampLabel',
        from_representation=from_representation,
    )


def timestamp_to_iso(timestamp, from_representation=None):
    """convert timestamp to TimestampISO

    ## Inputs
    - timestamp: Timestamp
    - from_representation: str representation name of input timestamp

    ## Returns
    - TimestampISO timestamp
    """
    return convert_timestamp(
        timestamp,
        to_representation='TimestampISO',
        from_representation=from_representation,
    )


def timestamp_to_datetime(
    timestamp: spec.Timestamp,
    from_representation: spec.TimestampRepresentation = None,
) -> spec.TimestampDatetime:
    """convert timestamp to TimestampDatetime

    ## Inputs
    - timestamp: Timestamp
    - from_representation: str representation name of input timestamp

    ## Returns
    - TimestampDatetime timestamp
    """
    return convert_timestamp(
        timestamp,
        to_representation='TimestampDatetime',
        from_representation=from_representation,
    )


#
# # specific conversion functions, from seconds
#


def timestamp_seconds_to_label(
    timestamp_seconds: spec.TimestampSecondsRaw,
) -> spec.TimestampLabel:
    """convert seconds to TimestampLabel"""
    dt = datetime.datetime.fromtimestamp(
        float(timestamp_seconds), datetime.timezone.utc
    )
    human_timestamp = dt.strftime(time_format)
    return human_timestamp


def timestamp_seconds_to_iso(
    timestamp_seconds: spec.TimestampSecondsRaw,
) -> spec.TimestampISO:
    """convert seconds to TimestampISO"""
    dt = datetime.datetime.fromtimestamp(
        float(timestamp_seconds), datetime.timezone.utc
    )
    iso_format = "%Y-%m-%dT%H:%M:%SZ"
    iso = dt.strftime(iso_format)
    return iso


def timestamp_seconds_to_datetime(
    timestamp_seconds: spec.TimestampSecondsRaw,
) -> spec.TimestampDatetime:
    """convert seconds to TimestampDatetime"""
    return datetime.datetime.fromtimestamp(
        float(timestamp_seconds), datetime.timezone.utc
    )


#
# # specific conversion functions, to seconds
#


def timestamp_label_to_seconds(
    timestamp_label: spec.TimestampLabel,
) -> spec.TimestampSecondsRaw:
    """convert TimestampLabel to seconds"""

    timestamp = timestamp_label

    if timestamp_label[-1] != 'Z' or len(timestamp_label) != 16:
        raise Exception('timestamp label not in format ' + str(time_format))

    dt = datetime.datetime(
        year=int(timestamp[:4]),
        month=int(timestamp[4:6]),
        day=int(timestamp[6:8]),
        hour=int(timestamp[9:11]),
        minute=int(timestamp[11:13]),
        second=int(timestamp[13:15]),
        tzinfo=datetime.timezone.utc,
    )
    return dt.timestamp()


def timestamp_iso_to_seconds(
    timestamp_iso: spec.TimestampISO,
) -> spec.TimestampSecondsRaw:
    """convert TimestampISO to seconds"""

    if '.' in timestamp_iso:
        iso_format = "%Y-%m-%dT%H:%M:%S.%fZ"
    else:
        iso_format = "%Y-%m-%dT%H:%M:%SZ"

    local_dt = datetime.datetime.strptime(timestamp_iso, iso_format)
    utc_dt = local_dt.replace(tzinfo=datetime.timezone.utc)
    seconds = utc_dt.timestamp()

    return seconds


def timestamp_datetime_to_seconds(
    timestamp_datetime: spec.TimestampDatetime,
) -> spec.TimestampSecondsRaw:
    """convert TimestampDatetime to seconds"""
    return timestamp_datetime.timestamp()


def timestamp_to_numerical(
    timestamp: spec.Timestamp,
) -> typing.Union[spec.TimestampSeconds, spec.TimestampSecondsPrecise]:
    if type(timestamp).__name__.startswith('int'):
        if isinstance(timestamp, typing.SupportsInt):
            return int(timestamp)
        else:
            raise Exception('bad type: ' + str(type(timestamp)))
    else:
        return timestamp_to_seconds_precise(timestamp)




from .. import timelength_utils


def detect_resolution(timestamps, use_n=None, outlier_rtol=0.5):
    """detect resolution of iterable of Timestamps

    ## Detection Algorithm
    - current algorithm: whether median dt value is close to a known value
    - dts that are +- 50% of this median are returned in 'outliers' key

    ## Inputs
    - timestamps: iterable of Timestamp
    - use_n: int or None, indicating number of timestamp values to use
        - smaller n will be faster but could be less representative
    - outlier_rtol: float of tolerance for detecting outliers
    """

    import numpy as np

    if len(timestamps) == 1:
        return None

    # preprocess
    if not isinstance(timestamps, np.ndarray):
        timestamps = np.array(timestamps)
    if use_n is not None:
        timestamps = timestamps[:use_n]

    # compute time deltas
    dts = timestamps[1:] - timestamps[:-1]
    median_dt = np.median(dts)
    outliers = {
        'small': dts[dts < 1 / (1 + outlier_rtol) * median_dt],
        'large': dts[dts > (1 + outlier_rtol) * median_dt],
        'outlier_rtol': outlier_rtol,
    }

    label = timelength_utils.timelength_seconds_to_label(median_dt)

    return {
        'label': label,
        'dts': dts,
        'use_n': use_n,
        'median_dt': median_dt,
        'outliers': outliers,
    }

